package ngjsvm

import (
	"context"
	"fmt"
	"github.com/GUAIK-ORG/go-snowflake/snowflake"
	"github.com/dop251/goja"
	"sync"
	"time"
)

const (
	JSRtSchedulerExceptionFunctionNotCallable = "provided function is not callable"
)

type jsRtSchedulerTicker struct {
	id       int64
	ticker   *time.Ticker
	callback func()
	ctx      context.Context
	cncl     context.CancelFunc
	onlyOnce bool
	delMet   func(tkid int64)
}

func newJsRtSchedulerTiker(
	id int64,
	interval time.Duration,
	callback func(),
	once bool,
	delMet func(tkid int64),
) *jsRtSchedulerTicker {
	ctx, cncl := context.WithCancel(context.Background())
	t := &jsRtSchedulerTicker{
		id:       id,
		ticker:   time.NewTicker(interval),
		callback: callback,
		ctx:      ctx,
		cncl:     cncl,
		onlyOnce: once,
		delMet:   delMet,
	}
	go t.tickLoop()
	return t
}

func (t *jsRtSchedulerTicker) tickLoop() {
TickLoop:
	for {
		select {
		case <-t.ctx.Done():
			break TickLoop
		case <-t.ticker.C:
			t.callback()
			if t.onlyOnce {
				break TickLoop
			}
			t.callback()
			break
		}
	}
	t.ticker.Stop()
	if t.delMet != nil {
		t.delMet(t.id)
	}
}

func (t *jsRtSchedulerTicker) Dispose() {
	if t.ticker != nil {
		t.ticker.Stop()
	}
	if t.cncl != nil {
		t.cncl()
		t.cncl = nil
	}
}

type JSRtScheduler struct {
	isReg      bool
	env        *JSEnv
	snowflaker *snowflake.Snowflake
	tickers    map[int64]*jsRtSchedulerTicker
	tickerLock sync.RWMutex
}

func (j *JSRtScheduler) clearAllTicks() {
	j.tickerLock.Lock()
	for k, v := range j.tickers {
		v.Dispose()
		delete(j.tickers, k)
	}
	j.tickerLock.Unlock()
}

func (j *JSRtScheduler) Dispose() {
	j.clearAllTicks()
}

func (j *JSRtScheduler) RegisterRt(name string, env *JSEnv) (goja.Value, error) {
	j.env = env
	sn, err := snowflake.NewSnowflake(0, 0)
	if err != nil {
		return nil, fmt.Errorf("failed create snowflaker: %v", err)
	}
	j.snowflaker = sn
	j.tickers = make(map[int64]*jsRtSchedulerTicker)
	fmap := map[string]interface{}{
		"NotCallableException": JSRtSchedulerExceptionFunctionNotCallable,
		"setInterval":          j.J_setInterval,
		"clearInterval":        j.J_clearInterval,
		"setTimeout":           j.J_setTimeout,
		"clearTimeout":         j.J_clearTimeout,
		"clearAll":             j.J_clearAll,
	}
	obj := j.env.BuildObject(name, fmap)
	j.isReg = true
	return obj, nil
}

func (j *JSRtScheduler) IsRegistered() bool {
	return j.isReg
}

func (j *JSRtScheduler) J_setInterval(call goja.FunctionCall) goja.Value {
	return j.setTicker(call, false)
}

func (j *JSRtScheduler) J_clearInterval(call goja.FunctionCall) goja.Value {
	return j.clearTick(call)
}

func (j *JSRtScheduler) J_setTimeout(call goja.FunctionCall) goja.Value {
	return j.setTicker(call, true)
}

func (j *JSRtScheduler) J_clearTimeout(call goja.FunctionCall) goja.Value {
	return j.clearTick(call)
}

func (j *JSRtScheduler) J_clearAll(call goja.FunctionCall) goja.Value {
	if !j.isReg {
		return goja.Undefined()
	}
	j.clearAllTicks()
	return goja.Undefined()
}

func (j *JSRtScheduler) setTicker(call goja.FunctionCall, once bool) goja.Value {
	if !j.isReg {
		return goja.Undefined()
	}
	if fn, ok := goja.AssertFunction(call.Argument(0)); ok {
		intvms := call.Argument(1).ToInteger()
		var args []goja.Value
		if len(call.Arguments) > 2 {
			args = append(args, call.Arguments[2:]...)
		}
		f := func() { fn(nil, args...) }
		intv := time.Duration(intvms) * time.Millisecond
		tkid := j.snowflaker.NextVal()
		tk := newJsRtSchedulerTiker(tkid, intv, f, once, j.disposeTicker)
		j.tickerLock.Lock()
		j.tickers[tkid] = tk
		j.tickerLock.Unlock()
		return j.env.ToValue(tkid)
	} else {
		j.env.ThrowException(JSRtSchedulerExceptionFunctionNotCallable)
	}
	return goja.Undefined()
}

func (j *JSRtScheduler) disposeTicker(tkid int64) {
	j.tickerLock.Lock()
	tk, ok := j.tickers[tkid]
	if ok {
		tk.Dispose()
		delete(j.tickers, tkid)
	}
	j.tickerLock.Unlock()
}

func (j *JSRtScheduler) clearTick(call goja.FunctionCall) goja.Value {
	if !j.isReg {
		return goja.Undefined()
	}
	tkidRaw := call.Argument(0)
	tkid := tkidRaw.ToInteger()
	j.disposeTicker(tkid)
	return goja.Undefined()
}
